#include "stdafx.h"

#include <algorithm>
#include <vector>

#include "tab4.h"
#include "ff3se.h"
#include "resource.h"
#include "character.h"
#include "options.h"
#include "choice.h"
#include "export.h"

Tab4 tab4;

char *spriteNames[numCharacters] = { "Terra", "Locke", "Cyan", "Shadow", "Edgar", "Sabin", "Celes", "Strago", "Relm", "Setzer", "Mog", "Gau", "Gogo", "Umaro",
	"Soldier", "Kappa", "Leo", "Banon", "Esper Terra", "Merchant Locke", "Ghost", "Kefka", "Gestahl", "Old Man", "Man", "Dog", "Maria", "Rich Man", "Draco",
	"Arvis", "Pilot", "Ultros", "Fancy Gau", "Dancer", "Chancellor", "Baram", "Old Woman", "Woman", "Boy", "Girl", "Bird", "Rachel", "Katarin",
	"Impressario", "Esper Elder", "Yura", "Siegfried", "Cid", "Maduin", "Bandit", "Vargas/Dadaluma", "Monster", "Narshe Guard", "Train Conductor",
	"Shopkeeper", "Faerie", "Wolf", "Dragon", "Fish", "Figaro Guard", "Darill", "Chupon", "Emperor's Servant", "Ramuh", "Figaro Guard Riding",
	"Celes in Chains", "Gau Sabin", "Gau Locke", "King Doma", "Number 128", "Magi Warrior A", "Skull Statue", "Ifrit", "Phantom", "Shiva",
	"Unicorn", "Bismark", "Carbuncl", "Shoat", "Owzer A", "Owzer B", "Blackjack", "Dead Figaro Guard", "Number 024", "Treasure Chest",
	"Magi Warrior B", "Atma", "Small Statue", "Flowers", "Envelope", "Plant", "Magicite", "Book", "Baby", "Question Mark", "Exclamation Point",
	"Slave Crown", "Weight", "Bird w/ Bandana", "Eyes in the Dark", "Bandana", "Nothing", "Bird Flying Left", "Bird Flying Up", "Big Sparkle",
	"Multiple Sparkles", "Small Sparkle", "Coin", "Rat", "Turtle", "Small Bird Flying Up", "Save Point", "Flame", "Explosion", "Tentacle A",
	"Tentacle B", "Big Switch", "Floor Switch", "Rock", "Crane Hook C", "Elevator", "Flying Esper Terra Front 1", "Flying Esper Terra Front 2",
	"Ending Terra Feet", "Diving Helmet", "Guardian A", "Guardian B", "Guardian C", "Crane Hook B", "Guardian D", "Guardian E", "Guardian F",
	"Crane Hook A", "Magitek Factory Machine", "Gate A", "Gate B", "Gate C", "Air Force", "Leo's Sword", "Magitek Train Car A", "Magitek Train Car B",
	"Magitek Train Car C", "Magitek Train Car D", "Crane A", "Crane B", "Crane C", "Chardarnook A", "Chardarnook B", "Chardarnook C", "Falcon A", "Falcon B",
	"Falcon C", "Flying Esper Terra Back", "Tritoch", "Odin", "Goddess A", "Doom A", "Poltergeist A", "Goddess B", "Goddess C", "Doom/Poltergeist B",
	"Doom/Poltergeist C", "Ending Terra Top Left", "Ending Terra Top Right", "Small Bird Flying Left"
};

/*************************************/
//Destructor for Tab1
Tab4::~Tab4(void)
{
	DeleteDC(spriteHdc);
	DeleteDC(editHdc);

	DeleteObject(spriteBMP);
	DeleteObject(editBMP);
}

/*************************************/
//Adds the list of character names to the combo box
void Tab4::addNames(void)
{
	int i;

	SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER, CB_RESETCONTENT, 0, 0);

	for (i = 0; i < numCharacters; i++)
	{
		SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER, CB_ADDSTRING, 0,
			(WPARAM)(LPCSTR)spriteNames[i]);
	}

	SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER, CB_SETCURSEL, 0, 0);

	charSelected = 0;
}

/*************************************/
//Handles the action of the brush button
//Sets the pixel x,y of the portrait palette entry selected
void Tab4::brushButton(int x, int y)
{
	m_changed = true;

	int height = character[charSelected].getPixelHeight();
	int width = character[charSelected].getPixelWidth();
	if (spriteSelected == numFullPics - 1)
	{
		std::swap(height, width);
	}
	if ((x >= width) || (y >= height))
		return;
	/*	if((spriteSelected == 2) || (spriteSelected == 5))
			x = 15 - x;*/

	character[charSelected].setPixel(spriteSelected, x, y, palEntrySelected);

	invalidatePics();
}

/*************************************/
//Changes which palette entry is selected
void Tab4::changePalEntry(BYTE num)
{
	palEntrySelected = num;
	sliderEnable();
	invalidateTabControl(IDC_PIC_CHARACTER_PALETTE);
}

/*************************************/
//Changes the palette used by the Charactert to the palette selected in the combo box
void Tab4::changePalette(void)
{
	int palSelected;

	m_changed = true;

	palSelected = SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER_PALETTE, CB_GETCURSEL, 0, 0);
	character[charSelected].changePalette(palSelected);

	invalidatePics();
}

/*************************************/
//Changes which color is transparent
//Swaps the currently selected color with color 0
void Tab4::changeTransparency(void)
{
	m_changed = true;

	character[charSelected].setTransparency(palEntrySelected);

	changePalEntry(0);

	invalidatePics();
}

/*************************************/
//Clears all the sprite images for the character
void Tab4::clearImage(void)
{
	char text[256];

	sprintf(text, "This will erase all sprites for %s", spriteNames[charSelected]);

	if (MessageBox(NULL, text, ff3seTitle, MB_OKCANCEL | MB_TASKMODAL) != IDOK)
		return;

	character[charSelected].clear();

	invalidatePics();
}

/*************************************/
//Copies another character's sprite(s)
void Tab4::copyImage(void)
{
	int r;
	int i;

	r = choiceDialog.show("Copy character sprite", "Copy only selected sprite",
		"Copy all sprites for this character");
	if (r == -1)
		return;
	if (r == 1)
		copyAll = -1;
	else
		copyAll = spriteSelected;

	m_changed = true;

	hDlg = CreateDialog(hInst, MAKEINTRESOURCE(IDD_COPY_IMAGE), hTab, ::copyCharacterProc);

	copyList.clear();
	SendDlgItemMessage(hDlg, IDC_COPY_LIST, LB_RESETCONTENT, 0, 0);
	for (i = 0; i < numCharacters; i++)
	{
		if (character[charSelected].canCopyFrom(character[i], copyAll))
		{
			copyList.push_back(i);
			SendDlgItemMessage(hDlg, IDC_COPY_LIST, LB_ADDSTRING, 0,
				(LPARAM)(LPCTSTR)spriteNames[i]);
		}
	}

	SetWindowText(hDlg, "Copy character sprite");
	BringWindowToTop(hDlg);
	ShowWindow(hDlg, SW_NORMAL);
	EnableWindow(hWnd, false);
}

/*************************************/
//Creates the tab
void Tab4::create(void)
{
	hTab = CreateDialog(hInst, MAKEINTRESOURCE(IDD_TAB4), ::hMainTab, tab4Proc);
	SetWindowPos(hTab, HWND_TOP, 5, 30, 680, 480, 0);
}

/*************************************/
//Updates HDCs for the character and editing pic with the new bitmap images
void Tab4::createHDCs(void)
{
	int xScroll, yScroll;

	xScroll = SendDlgItemMessage(hTab, IDC_HSCROLL_CHARACTER_EDIT, SBM_GETPOS, 0, 0);
	yScroll = SendDlgItemMessage(hTab, IDC_VSCROLL_CHARACTER_EDIT, SBM_GETPOS, 0, 0);

	createViewImage();

	spriteBmp.drawToDC(spriteHdc, &backBmp, 128, 144);

	character[charSelected].pics[spriteSelected].drawToLargeDC
	(magnification, editHdc, xScroll, yScroll, &back2Bmp, 256, 256);
}

/*************************************/
//Creates the view image which contains all sprites for a character
void Tab4::createViewImage(void)
{
	int i = 0;
	BYTE sprite;
	int x, y;
	int x2, y2;

	spriteBmp.createBlank(128, 144);

	//Set palettes
	for (i = 0; i < 16; i++)
	{
		spriteBmp.setColor(i, character[charSelected].pics[0].getColor(i));
	}

	int num_sprites = character[charSelected].getNumPics();
	int height = character[charSelected].getPixelHeight();
	int width = character[charSelected].getPixelWidth();

	//Create sprite view Bitmap
	for (y = 0; y < 144; y++)
	{
		y2 = y % height;
		for (x = 0; x < 128; x++)
		{
			x2 = x % width;
			sprite = (y / height) * 8 + (x / width);

			/*			if((sprite == 2) || (sprite == 5))
							x2 = 15 - x2;*/
			if (sprite > numFullPics || (sprite >= num_sprites && num_sprites < numFullPics))
			{
				spriteBmp.setPixel(x, y, 0);
			}
			else
			{
				if (sprite == numFullPics)
				{
					sprite--;
					x2 += 16;
				}
				if (sprite == numFullPics - 1)
				{
					spriteBmp.setPixel(x, y, character[charSelected].pics[numFullPics - 1].getPixel(x2, y2 - 8));
				}
				else
				{
					spriteBmp.setPixel(x, y, character[charSelected].pics[sprite].getPixel(x2, y2));
				}
			}
		}
	}

	if (spriteSelected >= num_sprites)
	{
		spriteSelected = 0;
	}
	//Draw border around selected sprite
	if (spriteSelected != -1)
	{
		int height = character[charSelected].getPixelHeight();
		int width = character[charSelected].getPixelWidth();
		x2 = (spriteSelected % 8) * width;
		y2 = (spriteSelected / 8) * height;

		spriteBmp.setColor(16, makeRGBTRIPLE(0xFF, 0xFF, 0xFF));

		if (spriteSelected == numFullPics - 1)
		{
			y2 += 8;
			std::swap(height, width);
		}
		for (i = 0; i < std::max(height, width); i++)
		{
			spriteBmp.setPixel(x2, y2 + i, 16);
			spriteBmp.setPixel(x2 + width - 1, y2 + i, 16);
			if (i < width)
			{
				spriteBmp.setPixel(x2 + i, y2, 16);
				spriteBmp.setPixel(x2 + i, y2 + height - 1, 16);
			}
		}
	}
}

/*************************************/
//Displays the preview of the file selected in the import dialog box
void Tab4::displayPreview(HWND hdlg)
{
	char filename[200];
	HDC picHdc, bmpHdc;
	HWND hPic, hText;
	Bitmap bmp;
	BYTE result;
	HBITMAP tempBMP;
	int size;

	size = CommDlg_OpenSave_GetSpec(GetParent(hdlg), filename, 200);
	if (size == 1)		//Exit if no file selected
	{
		SendDlgItemMessage(hdlg, IDC_TEXT_COLORS, WM_SETTEXT, (WPARAM)0, (LPARAM)(LPCTSTR)"");
		SendDlgItemMessage(hdlg, IDC_TEXT_SIZE, WM_SETTEXT, (WPARAM)0, (LPARAM)(LPCTSTR)"");
		return;
	}
	CommDlg_OpenSave_GetFilePath(GetParent(hdlg), filename, 200);
	hPic = GetDlgItem(hdlg, IDC_MONSPREVIEW);
	picHdc = GetDC(hPic);							//Create an HDC for the preview picture control

	result = bmp.loadFile(filename, 16, 128, 128);
	if (result != 0)
	{
		backBmp.drawToDC(picHdc);

		//Set text message to display reason image can't be previewed
		SendDlgItemMessage(hdlg, IDC_TEXT_COLORS, WM_SETTEXT, (WPARAM)0, (LPARAM)(LPCTSTR)"");
		if (result == 1)
		{
			SendDlgItemMessage(hdlg, IDC_TEXT_SIZE, WM_SETTEXT, (WPARAM)0,
				(LPARAM)(LPCTSTR)"Image has over 16 colors");
		}
		else if (result == 2)
		{
			SendDlgItemMessage(hdlg, IDC_TEXT_SIZE, WM_SETTEXT, (WPARAM)0,
				(LPARAM)(LPCTSTR)"Image is larger than 128x128");
		}
		else if (result == 3)
		{
			SendDlgItemMessage(hdlg, IDC_TEXT_SIZE, WM_SETTEXT, (WPARAM)0,
				(LPARAM)(LPCTSTR)"Error opening image");
		}
		else if (result == 4)
		{
			SendDlgItemMessage(hdlg, IDC_TEXT_SIZE, WM_SETTEXT, (WPARAM)0,
				(LPARAM)(LPCTSTR)"Not a valid image file");
		}
	}
	else
	{
		bmpHdc = CreateCompatibleDC(screenHdc);
		tempBMP = CreateCompatibleBitmap(screenHdc, 128, 128);
		SelectObject(bmpHdc, tempBMP);
		bmp.drawToDC(bmpHdc, &backBmp, 128, 128);

		BitBlt(picHdc, 0, 0, 128, 128, bmpHdc, 0, 0, SRCCOPY);

		hText = GetDlgItem(hdlg, IDC_TEXT_COLORS);

		SendDlgItemMessage(hdlg, IDC_TEXT_COLORS, WM_SETTEXT, (WPARAM)0, (LPARAM)(LPCTSTR)"");
		SendDlgItemMessage(hdlg, IDC_TEXT_SIZE, WM_SETTEXT, (WPARAM)0, (LPARAM)(LPCTSTR)"");
	}


	DeleteDC(bmpHdc);
	DeleteDC(picHdc);
	DeleteObject(tempBMP);
}

/*************************************/
//Redraws the paint buttons
void Tab4::drawPaintButtons(WPARAM wParam, LPARAM lParam)
{
	HBITMAP BMP;
	HDC bufferHdc;

	switch (((LPDRAWITEMSTRUCT)lParam)->CtlID)
	{
	case IDC_BUTTON_CHARACTER_BRUSH:
		if (buttonSelected == 0)
			BMP = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_BRUSH2), IMAGE_BITMAP, 42, 42, LR_DEFAULTCOLOR);
		else
			BMP = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_BRUSH), IMAGE_BITMAP, 42, 42, LR_DEFAULTCOLOR);
		bufferHdc = CreateCompatibleDC(screenHdc);
		SelectObject(bufferHdc, BMP);
		BitBlt(((LPDRAWITEMSTRUCT)lParam)->hDC, 0, 0, 42, 42, bufferHdc, 0, 0, SRCCOPY);
		break;
	case IDC_BUTTON_CHARACTER_ERASE:
		if (buttonSelected == 1)
			BMP = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_ERASE2), IMAGE_BITMAP, 42, 42, LR_DEFAULTCOLOR);
		else
			BMP = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_ERASE), IMAGE_BITMAP, 42, 42, LR_DEFAULTCOLOR);
		bufferHdc = CreateCompatibleDC(screenHdc);
		SelectObject(bufferHdc, BMP);
		BitBlt(((LPDRAWITEMSTRUCT)lParam)->hDC, 0, 0, 42, 42, bufferHdc, 0, 0, SRCCOPY);
		break;
	case IDC_BUTTON_CHARACTER_GRAB:
		if (buttonSelected == 2)
			BMP = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_GRAB2), IMAGE_BITMAP, 42, 42, LR_DEFAULTCOLOR);
		else
			BMP = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_GRAB), IMAGE_BITMAP, 42, 42, LR_DEFAULTCOLOR);
		bufferHdc = CreateCompatibleDC(screenHdc);
		SelectObject(bufferHdc, BMP);
		BitBlt(((LPDRAWITEMSTRUCT)lParam)->hDC, 0, 0, 42, 42, bufferHdc, 0, 0, SRCCOPY);
		break;
	}
	DeleteDC(bufferHdc);
	DeleteObject(BMP);
}

/*************************************/
void Tab4::drawTab(void)
{
	HDC picHdc;
	HDC bufferHdc;
	HWND hPic;
	HBITMAP bufferBMP;
	PAINTSTRUCT ps;
	RECT r, r2;
	int x, y, i, j;
	long color;
	int g;
	int maxX, maxY;

	if (!hTab)		//Exit if attempting to draw before the tab is fully initiliazed
		return;

	createHDCs();

	GetUpdateRect(hTab, &r, false);

	ps.fErase = false;
	ps.rcPaint = r;
	GetWindowRect(hTab, &r2);

	r.bottom += r2.top;		//Convert update rect to screen coordinates
	r.left += r2.left;
	r.right += r2.left;
	r.top += r2.top;

	//Draw sprite pic
	hPic = GetDlgItem(hTab, IDC_PIC_CHARACTER_SPRITE);

	GetWindowRect(hPic, &r2);

	if (overlapRect(r, r2))			//Check if sprite pic needs to be updated
	{
		picHdc = GetDC(hPic);
		ps.hdc = picHdc;

		BeginPaint(hTab, &ps);

		BitBlt(picHdc, 0, 0, 128, 144, spriteHdc, 0, 0, SRCCOPY);

		EndPaint(hTab, &ps);
		ReleaseDC(hPic, picHdc);
	}

	//Draw zoomed in character editing pic
	hPic = GetDlgItem(hTab, IDC_PIC_EDITCHARACTER);

	GetWindowRect(hPic, &r2);

	if (overlapRect(r, r2))			//Check if edit character pic needs to be updated
	{
		picHdc = GetDC(hPic);
		ps.hdc = picHdc;

		BeginPaint(hTab, &ps);

		int height = character[charSelected].getPixelHeight();
		int width = character[charSelected].getPixelWidth();

		if (spriteSelected == numFullPics - 1)
		{
			std::swap(height, width);
		}
		maxX = width * magnification;
		if (maxX > 256)
			maxX = 256;
		maxY = height * magnification;
		if (maxY > 256)
			maxY = 256;

		g = options.drawGrid(magnification);
		if (g == 1)		//Show pixel grid
		{
			for (i = 0; i <= 256; i += magnification)
			{
				for (j = 0; j <= 256; j++)
				{
					if ((i <= maxX) && (j <= maxY))
						SetPixel(editHdc, i, j, 0x888888);
					if ((i <= maxY) && (j <= maxX))
						SetPixel(editHdc, j, i, 0x888888);
				}
			}
		}
		if (g == 2)		//Show tile grid
		{
			int xoff, yoff;
			int xScroll, yScroll;

			xScroll = SendDlgItemMessage(hTab, IDC_HSCROLL_MONSTER_EDIT, SBM_GETPOS, 0, 0);
			yScroll = SendDlgItemMessage(hTab, IDC_VSCROLL_MONSTER_EDIT, SBM_GETPOS, 0, 0);

			xoff = (xScroll % 8) * magnification;
			yoff = (yScroll % 8) * magnification;

			for (i = 0; i <= 256; i += magnification * 8)
			{
				for (j = 0; j <= 256; j++)
				{
					x = i - xoff;
					if (x < 0)
						x += (256 - (256 % (magnification * 8))) + (magnification * 8);
					y = j;
					if ((x <= maxX) && (y <= maxY))
						SetPixel(editHdc, x, y, 0x888888);
					y = i - yoff;
					if (y < 0)
						y += (256 - (256 % (magnification * 8))) + (magnification * 8);
					x = j;
					if ((x <= maxX) && (y <= maxY))
						SetPixel(editHdc, x, y, 0x888888);
				}
			}
		}

		BitBlt(picHdc, 0, 0, 256, 256, editHdc, 0, 0, SRCCOPY);

		EndPaint(hTab, &ps);
		ReleaseDC(hPic, picHdc);
	}

	//Draw palette pic
	hPic = GetDlgItem(hTab, IDC_PIC_CHARACTER_PALETTE);

	GetWindowRect(hPic, &r2);

	if (overlapRect(r, r2))			//Check if palette pic needs to be updated
	{
		picHdc = GetDC(hPic);
		ps.hdc = picHdc;

		//Create a buffer to draw the palette on
		bufferHdc = CreateCompatibleDC(screenHdc);
		bufferBMP = CreateCompatibleBitmap(screenHdc, 288, 30);
		SelectObject(bufferHdc, bufferBMP);

		//Draw the palette with a border around each color
		for (i = 0; i < 16; i++)
		{
			color = RGBtoCOLORREF(charPals[character[charSelected].getPalNum()][i]);
			for (y = 0; y < 30; y++)
			{
				if (palEntrySelected == i)
					SetPixel(bufferHdc, i * 18, y, 0xFFFFFF);
				else
					SetPixel(bufferHdc, i * 18, y, 0);
				for (x = 1; x < (18 - 1); x++)
				{
					if ((y == 0) || (y == 29))
						if (palEntrySelected == i)
							SetPixel(bufferHdc, x + i * 18, y, 0xFFFFFF);
						else
							SetPixel(bufferHdc, x + i * 18, y, 0);
					else
						SetPixel(bufferHdc, x + i * 18, y, color);
				}
				if (palEntrySelected == i)
					SetPixel(bufferHdc, x + i * 18, y, 0xFFFFFF);
				else
					SetPixel(bufferHdc, x + i * 18, y, 0);
			}
		}

		//Draw buffer to screen
		BeginPaint(hTab, &ps);

		BitBlt(picHdc, 0, 0, 288, 30, bufferHdc, 0, 0, SRCCOPY);

		EndPaint(hTab, &ps);
		ReleaseDC(hPic, picHdc);
		DeleteDC(bufferHdc);
		DeleteObject(bufferBMP);
	}

	updateControls();
}

/*************************************/
//Handles the action of the erase button
//Sets the pixel x,y of the portrait to color 0 (transparent)
void Tab4::eraseButton(int x, int y)
{
	m_changed = true;

	int height = character[charSelected].getPixelHeight();
	int width = character[charSelected].getPixelWidth();

	if (spriteSelected == numFullPics - 1)
	{
		std::swap(height, width);
	}

	if ((x >= width) || (y >= height))
		return;

	// ?
	//if ((spriteSelected = 2) || (spriteSelected == 5))
	//	x = 15 - x;

	character[charSelected].setPixel(spriteSelected, x, y, palEntrySelected);

	invalidatePics();
}

/*************************************/
void Tab4::exportImage(void)
{
	char filename[1024];
	int fileType;
	int oldSel;

	sprintf(filename, "%s.png", spriteNames[charSelected]);
	fileType = 2;

	if (exportDialog.show("Export character sprites", filename, &fileType, true))
	{
		oldSel = spriteSelected;
		spriteSelected = -1;		//Make sure no border around the selected
		
		createViewImage();			//sprite is in exported image
		switch (fileType)
		{
		case 0:
			spriteBmp.saveBMP(filename, 128, 144);
			break;
		case 1:
			spriteBmp.savePCX(filename, 128, 144);
			break;
		case 2:
			spriteBmp.savePNG(filename, 128, 144);
			break;
		case 3:
			auto handle = _open(filename, _O_WRONLY | _O_BINARY | _O_TRUNC | _O_CREAT, _S_IREAD | _S_IWRITE);
			if (handle == -1)
			{
				MessageBox(NULL, "Error saving sprite data", ff3seTitle, MB_OK | MB_ICONWARNING | MB_TASKMODAL);
				return;
			}
			character[charSelected].save(handle);
			_close(handle);
		}
		spriteSelected = oldSel;
		createViewImage();
	}
}

/*************************************/
void Tab4::flipImage(int flip)
{
	character[charSelected].flipImage(flip);
	invalidatePics();
}

/*************************************/
//Handles the scroll bars for the editing pic
void Tab4::handleMagScroll(int val, HWND hScroll, int pos)
{
	switch (val)
	{
	case SB_LINELEFT:
		pos = GetScrollPos(hScroll, SB_CTL) - 1;
		SetScrollPos(hScroll, SB_CTL, pos, true);
		break;
	case SB_LINERIGHT:
		pos = GetScrollPos(hScroll, SB_CTL) + 1;
		SetScrollPos(hScroll, SB_CTL, pos, true);
		break;
	case SB_PAGELEFT:
		pos = GetScrollPos(hScroll, SB_CTL) - 10;
		SetScrollPos(hScroll, SB_CTL, pos, true);
		break;
	case SB_PAGERIGHT:
		pos = GetScrollPos(hScroll, SB_CTL) + 10;
		SetScrollPos(hScroll, SB_CTL, pos, true);
		break;
	case SB_THUMBTRACK:
		SetScrollPos(hScroll, SB_CTL, pos, true);
	}

}

/*************************************/
//Determines where the mouse was clicked and handles any operations that need to be done
//Changes the palette selction if the palette bar is clicked on.
void Tab4::handleMouseClick(int x, int y)
{
	int tabx, taby;
	HWND hPal, hEdit, hView;
	BYTE c;
	BYTE s;
	RECT r, r2;

	GetWindowRect(hTab, &r);

	tabx = r.left;
	taby = r.top;

	x += tabx;
	y += taby;

	//Check if mouse was clicked on sprite view pic
	hView = GetDlgItem(hTab, IDC_PIC_CHARACTER_SPRITE);
	GetWindowRect(hView, &r);

	if (inRect(x, y, r))
	{
		x -= r.left;
		y -= r.top;
		if (y > 144)
			y = 144;

		int height = character[charSelected].getPixelHeight();
		int width = character[charSelected].getPixelWidth();
		int numSprites = character[charSelected].getNumPics();
		s = (x / width) + (y / height) * 8;
		if (s == numFullPics && x % width < 8)
		{
			s--;
		}
		if (s < numSprites && (character[charSelected].type != Character::NO_RIDING || s < numPreRidingPics || s == numFullPics - 1))
			spriteSelected = s;

		invalidatePics();

		x += r.left;
		y += r.top;
	}

	//Check if the mouse was clicked on the palette bar
	hPal = GetDlgItem(hTab, IDC_PIC_CHARACTER_PALETTE);
	GetWindowRect(hPal, &r);

	if (inRect(x, y, r))
	{
		changePalEntry((x - r.left) / 18);
	}

	//Check if the mouse was clicked on the editing pic and fill was selected

	hEdit = GetDlgItem(hTab, IDC_PIC_EDITCHARACTER);
	GetWindowRect(hEdit, &r);

	if (inRect(x, y, r))
	{
		GetWindowRect(hEdit, &r2);

		//Find what pixel was clicked on
		y -= r2.top;
		x -= r2.left;

		x /= magnification;
		y /= magnification;

		x += SendDlgItemMessage(hTab, IDC_HSCROLL_CHARACTER_EDIT, SBM_GETPOS, 0, 0);
		y += SendDlgItemMessage(hTab, IDC_VSCROLL_CHARACTER_EDIT, SBM_GETPOS, 0, 0);

		if (buttonSelected == 0)
		{
			brushButton(x, y);
		}
		if (buttonSelected == 1)
		{
			eraseButton(x, y);
		}
		if (buttonSelected == 2)
		{
			c = character[charSelected].pics[spriteSelected].getPixel(x, y);

			changePalEntry(c);
		}
	}

}

/*************************************/
//Determines where the mouse is currently at
//Down is set if the left mouse button is pressed
//Handles updating the portrait image, if the editing pic is clicked on
void Tab4::handleMouseMove(int x, int y, bool down)
{

	if (!down)			//Don't edit the pic unless the left mouse button is pressed
		return;

	int tabx, taby;
	HWND hEdit;
	BYTE c;
	RECT r, r2;

	GetWindowRect(hTab, &r);

	tabx = r.left;
	taby = r.top;

	x += tabx;
	y += taby;

	//Check if the mouse was clicked on the editing pic

	hEdit = GetDlgItem(hTab, IDC_PIC_EDITCHARACTER);
	GetWindowRect(hEdit, &r);

	if (inRect(x, y, r))
	{
		GetWindowRect(hEdit, &r2);

		//Find what pixel was clicked on
		y -= r2.top;
		x -= r2.left;

		x /= magnification;
		y /= magnification;

		x += SendDlgItemMessage(hTab, IDC_HSCROLL_CHARACTER_EDIT, SBM_GETPOS, 0, 0);
		y += SendDlgItemMessage(hTab, IDC_VSCROLL_CHARACTER_EDIT, SBM_GETPOS, 0, 0);

		if (buttonSelected == 0)
		{
			brushButton(x, y);
		}
		if (buttonSelected == 1)
		{
			eraseButton(x, y);
		}
		if (buttonSelected == 2)
		{
			c = character[charSelected].pics[spriteSelected].getPixel(x, y);

			changePalEntry(c);
		}
	}
}

/*************************************/
//Handles functions for the palette sliders
void Tab4::handleSliders(WPARAM wParam, LPARAM lParam)
{
	HWND hSlider;
	BYTE val;
	int i;

	m_changed = true;

	hSlider = GetDlgItem(hTab, IDC_SLIDER_CHARACTER_RED);

	if ((HWND)lParam == hSlider)
	{
		val = (BYTE)SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
		charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtRed = val * 8;
		for (i = 0; i < numCharacters; i++)
			character[i].updatePalette();
		invalidatePics();
	}

	hSlider = GetDlgItem(hTab, IDC_SLIDER_CHARACTER_GREEN);

	if ((HWND)lParam == hSlider)
	{
		val = (BYTE)SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
		charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtGreen = val * 8;
		for (i = 0; i < numCharacters; i++)
			character[i].updatePalette();
		invalidatePics();
	}

	hSlider = GetDlgItem(hTab, IDC_SLIDER_CHARACTER_BLUE);

	if ((HWND)lParam == hSlider)
	{
		val = (BYTE)SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
		charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtBlue = val * 8;
		for (i = 0; i < numCharacters; i++)
			character[i].updatePalette();
		invalidatePics();
	}

}

/*************************************/
//Handles imports a new tileset of sprite set
//Called when import is selected from the Image menu
void Tab4::import(void)
{
	OPENFILENAME of;

	char filename[256];
	char *title;

	title = (char *)malloc(50);
	strcpy(title, "Import Character sprites");

	ZeroMemory(filename, 256);

	of.Flags = OFN_ENABLETEMPLATE | OFN_EXPLORER | OFN_HIDEREADONLY
		| OFN_FILEMUSTEXIST | OFN_ENABLEHOOK;
	of.hInstance = hInst;
	of.hwndOwner = hTab;
	of.lCustData = 0;
	of.lpfnHook = ::tab4ImportProc;
	of.lpstrCustomFilter = NULL;
	of.lpstrDefExt = NULL;
	of.lpstrFile = filename;
	of.lpstrFileTitle = NULL;
	of.lpstrFilter = "Supported images formats (*.png, *.bmp, *.pcx)\000*.png;*.bmp;*.pcx\000*.bin\000*.bin\000All files (*.*)\000*.*\000\000";
	of.lpstrInitialDir = NULL;
	of.lpstrTitle = title;
	of.lpTemplateName = MAKEINTRESOURCE(IDD_IMPORT_MONS);
	of.lStructSize = sizeof(OPENFILENAME);
	of.nFileExtension = 0;
	of.nFileOffset = 0;
	of.nFilterIndex = 0;
	of.nMaxCustFilter = 0;
	of.nMaxFile = 256;
	of.nMaxFileTitle = 256;

	GetOpenFileName(&of);

	m_changed = true;
}

/*************************************/
void Tab4::initTab(void)
{
	int i;
	char temp[20];
	char temp2[20];

	spriteBMP = CreateCompatibleBitmap(screenHdc, 128, 144);
	editBMP = CreateCompatibleBitmap(screenHdc, 256, 256);

	spriteHdc = CreateCompatibleDC(screenHdc);
	editHdc = CreateCompatibleDC(screenHdc);

	SelectObject(spriteHdc, spriteBMP);
	SelectObject(editHdc, editBMP);

	addNames();

	spriteBmp.createBlank(128, 144);

	charSelected = 0;
	spriteSelected = 0;

	//Set ranges for color sliders
	SendDlgItemMessage(hTab, IDC_SLIDER_CHARACTER_RED, TBM_SETRANGE, true, MAKELONG(0, 32));
	SendDlgItemMessage(hTab, IDC_SLIDER_CHARACTER_GREEN, TBM_SETRANGE, true, MAKELONG(0, 32));
	SendDlgItemMessage(hTab, IDC_SLIDER_CHARACTER_BLUE, TBM_SETRANGE, true, MAKELONG(0, 32));

	//Add list of palettes to combo box
	SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER_PALETTE, CB_RESETCONTENT, 0, 0);
	for (i = 0; i < 32; i++)
	{
		strcpy(temp, "Palette #");
		strcat(temp, _itoa(i, temp2, 10));
		SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER_PALETTE, CB_ADDSTRING, 0,
			(LPARAM)(LPCTSTR)temp);
	}

	magnification = options.getTab4Mag();

	SendDlgItemMessage(hTab, IDC_UPDOWN_CHARACTER_MAGNIFICATION, UDM_SETRANGE, 0, (LPARAM)MAKELONG((short)20, (short)2));
	SendDlgItemMessage(hTab, IDC_UPDOWN_CHARACTER_MAGNIFICATION, UDM_SETPOS, 0, (LPARAM)MAKELONG((short)magnification, 0));

	palEntrySelected = 0;

	newSelection();
}

/*************************************/
//Invalidates the view pic,the editing pic, and the palette bar
//Calls drawTab to immediately update and redraw the tab
void Tab4::invalidatePics(void)
{
	invalidateTabControl(IDC_PIC_CHARACTER_SPRITE);
	invalidateTabControl(IDC_PIC_EDITCHARACTER);
	invalidateTabControl(IDC_PIC_CHARACTER_PALETTE);

	drawTab();
}

/*************************************/
//handles the selection of a new character's portrait to edit
void Tab4::newSelection(void)
{
	int i;
	HWND hCtrl;
	bool enable = true;

	charSelected = (BYTE)SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER, CB_GETCURSEL, 0, 0);

	character[charSelected].createBitmaps();

	//Disable changing palette for Esper Terra
	if(charSelected == 18)	
	{
		enable = false;
	}

	hCtrl = GetDlgItem(hTab, IDC_COMBO_CHARACTER_PALETTE);
	EnableWindow(hCtrl, enable);
	sliderEnable();

	i = character[charSelected].getPalNum();
	SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER_PALETTE, CB_SETCURSEL, i, 0);

	invalidatePics();
}

/*************************************/
bool Tab4::save_internal(char *filename)
{
	return saveCharacters(filename);
}

/*************************************/
//Selected the nth paint button
void Tab4::selectButton(BYTE n)
{
	buttonSelected = n;
	invalidateTab();
}

/*************************************/
//If the mouse is over the editing pic, changes the cursor to the appropriate cursor
void Tab4::setCursor(void)
{
	POINT mousePos;
	HWND hEdit;
	RECT r;

	GetCursorPos(&mousePos);

	hEdit = GetDlgItem(hTab, IDC_PIC_EDITCHARACTER);
	GetWindowRect(hEdit, &r);

	if (inRect(mousePos.x, mousePos.y, r))
	{
		switch (buttonSelected)
		{
		case 0:
			SetCursor(brushCursor);
			break;
		case 1:
			SetCursor(eraseCursor);
			break;
		case 2:
			SetCursor(grabCursor);
			break;
		}
	}
	else
		SetCursor(LoadCursor(NULL, IDC_ARROW));
}

/*************************************/
//If the selected palette entry is 0 disables the palette sliders, otherwise enables them
void Tab4::sliderEnable(void)
{
	HWND hSlider;
	bool enable = true;

	if ((palEntrySelected == 0) || (charSelected == 18))
	{
		enable = false;
	}
	hSlider = GetDlgItem(hTab, IDC_SLIDER_CHARACTER_RED);
	EnableWindow(hSlider, enable);
	hSlider = GetDlgItem(hTab, IDC_SLIDER_CHARACTER_GREEN);
	EnableWindow(hSlider, enable);
	hSlider = GetDlgItem(hTab, IDC_SLIDER_CHARACTER_BLUE);
	EnableWindow(hSlider, enable);
	hSlider = GetDlgItem(hTab, IDC_CHECK_CHARACTER_TRANSPARENT);
	EnableWindow(hSlider, enable);
	SendMessage(hSlider, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
}

/*************************************/
//Updates all the combo boxes, sliders, etc.
void Tab4::updateControls(void)
{
	char temp[10];
	char mag[20];
	int maxScroll;

	int i;

	SendDlgItemMessage(hTab, IDC_SLIDER_CHARACTER_RED, TBM_SETPOS, true,
		charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtRed / 8);
	SendDlgItemMessage(hTab, IDC_SLIDER_CHARACTER_GREEN, TBM_SETPOS, true,
		charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtGreen / 8);
	SendDlgItemMessage(hTab, IDC_SLIDER_CHARACTER_BLUE, TBM_SETPOS, true,
		charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtBlue / 8);

	SendDlgItemMessage(hTab, IDC_TEXT_CHARACTER_REDVAL, WM_SETTEXT, (WPARAM)0,
		(LPARAM)(LPCTSTR)_itoa(charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtRed / 8,
			temp, 10));
	SendDlgItemMessage(hTab, IDC_TEXT_CHARACTER_GREENVAL, WM_SETTEXT, (WPARAM)0,
		(LPARAM)(LPCTSTR)_itoa(charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtGreen / 8,
			temp, 10));
	SendDlgItemMessage(hTab, IDC_TEXT_CHARACTER_BLUEVAL, WM_SETTEXT, (WPARAM)0,
		(LPARAM)(LPCTSTR)_itoa(charPals[character[charSelected].getPalNum()][palEntrySelected].rgbtBlue / 8,
			temp, 10));

	i = character[charSelected].getPalNum();
	SendDlgItemMessage(hTab, IDC_COMBO_CHARACTER_PALETTE, CB_SETCURSEL, i, 0);

	magnification = (BYTE)SendDlgItemMessage(hTab, IDC_UPDOWN_CHARACTER_MAGNIFICATION, UDM_GETPOS, 0, 0);

	if (magnification == 0)
		magnification = 1;

	options.setTab4Mag(magnification);

	sprintf(mag, "magnification: x%d", magnification);
	SendDlgItemMessage(hTab, IDC_TEXT_CHARACTER_MAGNIFICATION, WM_SETTEXT, 0, (LPARAM)(LPCTSTR)(mag));

	maxScroll = 16 - (256 / magnification);

	if (maxScroll < 0)
	{
		maxScroll = 0;
	}

	HWND hScroll = GetDlgItem(hTab, IDC_HSCROLL_CHARACTER_EDIT);
	SetScrollRange(hScroll, SB_CTL, 0, maxScroll, true);
	if (maxScroll > 0)
		ShowWindow(hScroll, true);
	else
		ShowWindow(hScroll, false);

	maxScroll = 24 - (256 / magnification);

	if (maxScroll < 0)
	{
		maxScroll = 0;
	}

	hScroll = GetDlgItem(hTab, IDC_VSCROLL_CHARACTER_EDIT);
	SetScrollRange(hScroll, SB_CTL, 0, maxScroll, true);
	if (maxScroll > 0)
		ShowWindow(hScroll, true);
	else
		ShowWindow(hScroll, false);
}

/*************************************/
UINT Tab4::importProc(HWND hdlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (message == WM_NOTIFY)
	{
		switch (((LPOFNOTIFY)lParam)->hdr.code)
		{
		case CDN_SELCHANGE:
			displayPreview(hdlg);
			return 1;
			break;
		case CDN_FILEOK:
			const char* filename = ((LPOFNOTIFY)lParam)->lpOFN->lpstrFile;
			int len = strlen(filename);

			if (len > 4 && (tolower(filename[len - 4]) == '.') &&
				(tolower(filename[len - 3]) == 'b') &&
				(tolower(filename[len - 2]) == 'i') &&
				(tolower(filename[len - 1]) == 'n'))
			{
				auto spriteHandle = _open(filename, O_RDWR | O_BINARY);
				if (spriteHandle == -1)
				{
					MessageBox(NULL, "Error opening file", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
					SetWindowLong(hdlg, DWL_MSGRESULT, 1);
					return 1;
				}

				int len = character[charSelected].getNumTiles() * 32;
				std::vector<BYTE> temp(len);

				_read(spriteHandle, temp.data(), len);

				character[charSelected].load(temp.data());
				_close(spriteHandle);
				invalidatePics();
				return 0;
			}
			Bitmap newBMP;

			switch (newBMP.loadFile(((LPOFNOTIFY)lParam)->lpOFN->lpstrFile, 16, 128, 144))
			{
			case 0:
				character[charSelected].setImage(newBMP);
				invalidatePics();
				return 0;
				break;
			case 1:
				MessageBox(NULL, "Image has more than 16 colors", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
				break;
			case 2:
				MessageBox(NULL, "Image is larger than 128x144", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
				break;
			case 3:
				MessageBox(NULL, "Error opening file", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
				break;
			case 4:
				MessageBox(NULL, "File does not appear to be a valid image file", ff3seTitle, MB_OK | MB_ICONERROR | MB_TASKMODAL);
				break;
			}
			SetWindowLong(hdlg, DWL_MSGRESULT, 1);
			return 1;
			break;
		}
	}
	return 0;
}

/*************************************/
bool Tab4::copyImageProc(HWND hdlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (message == WM_COMMAND)
	{
		switch (LOWORD(wParam))
		{
		case IDC_COPY_LIST:
			if (HIWORD(wParam) != LBN_DBLCLK)
				break;
		case IDC_COPY_OK:
			HWND hList;
			int image;

			hList = GetDlgItem(hdlg, IDC_COPY_LIST);
			image = SendDlgItemMessage(hdlg, IDC_COPY_LIST, LB_GETCURSEL, 0, 0);

			character[charSelected].copyImage(copyList[image], copyAll);

			invalidatePics();

			BringWindowToTop(hWnd);
			EnableWindow(hWnd, true);
			DestroyWindow(hdlg);
			return true;
			break;
		case IDC_COPY_CANCEL:
			BringWindowToTop(hWnd);
			EnableWindow(hWnd, true);
			DestroyWindow(hdlg);
			hDlg = NULL;
			return true;
			break;
		}
	}
	return false;
}

/*************************************/
BOOL Tab4::tabProc(HWND hdlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hScroll;

	switch (message)
	{
	case WM_CLOSE:
		hTab = NULL;
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_BUTTON_EXIT:
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			return true;
			break;
		case IDC_COMBO_CHARACTER:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				newSelection();
			}
			return true;
			break;
		case IDC_BUTTON_CHARACTER_BRUSH:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				selectButton(0);
			}
			return true;
			break;
		case IDC_BUTTON_CHARACTER_ERASE:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				selectButton(1);
			}
			return true;
			break;
		case IDC_BUTTON_CHARACTER_GRAB:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				selectButton(2);
			}
			return true;
			break;
		case IDC_COMBO_CHARACTER_PALETTE:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				changePalette();
			}
			return true;
			break;
		case IDC_CHECK_CHARACTER_TRANSPARENT:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				changeTransparency();
			}
			return true;
			break;
		case IDC_BUTTON_CHARACTER_CLEAR:
			clearImage();
			return true;
			break;
		case IDC_BUTTON_CHARACTER_COPY:
			copyImage();
			return true;
			break;
		case IDC_BUTTON_CHARACTER_IMPORT:
			import();
			return true;
			break;
		}
		break;
	case WM_DRAWITEM:
		drawPaintButtons(wParam, lParam);
		break;
	case  WM_HSCROLL:
		hScroll = GetDlgItem(hTab, IDC_HSCROLL_CHARACTER_EDIT);
		if ((HWND)lParam == hScroll)
		{
			handleMagScroll(LOWORD(wParam), (HWND)lParam, HIWORD(wParam));
			invalidatePics();
		}
		else
		{
			handleSliders(wParam, lParam);
		}
		return false;
		break;
	case WM_LBUTTONDOWN:
		handleMouseClick(LOWORD(lParam), HIWORD(lParam));
		return true;
		break;
	case WM_MOUSEMOVE:
		handleMouseMove(LOWORD(lParam), HIWORD(lParam), wParam & MK_LBUTTON);
		return true;
		break;
	case WM_PAINT:
		drawTab();
		return false;
		break;
	case WM_SETCURSOR:
		setCursor();
		return true;
		break;
	case WM_VSCROLL:
		hScroll = GetDlgItem(hTab, IDC_VSCROLL_CHARACTER_EDIT);
		if ((HWND)lParam == hScroll)
		{
			handleMagScroll(LOWORD(wParam), (HWND)lParam, HIWORD(wParam));
		}
		invalidatePics();
		return false;
		break;
	}
	return false;
}

/*************************************/
BOOL CALLBACK tab4Proc(HWND hdlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	return tab4.tabProc(hdlg, message, wParam, lParam);
}

/*************************************/
BOOL CALLBACK copyCharacterProc(HWND hdlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	return tab4.copyImageProc(hdlg, message, wParam, lParam);
}

/*************************************/
UINT CALLBACK tab4ImportProc(HWND hdlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	return tab4.importProc(hdlg, message, wParam, lParam);
}

